from typing import Dict, Final, Iterable, Optional, Union

from smllib.sml import SmlListEntry

from sml2mqtt.mqtt import MqttObj
from sml2mqtt.value.filter import FILTER_OBJ
from sml2mqtt.value.transformations import TRANSFORM_OBJ
from sml2mqtt.value.workarounds import WORKAROUND_OBJ


class SmlValue:
    def __init__(self, device: str, obis: str, mqtt: MqttObj,
                 workarounds: Iterable[WORKAROUND_OBJ],
                 transformations: Iterable[TRANSFORM_OBJ],
                 filters: Iterable[FILTER_OBJ]):

        self.device_id: Final = device
        self.obis: Final = obis
        self.mqtt: Final = mqtt

        self.sml_value: Optional[SmlListEntry] = None
        self.last_value: Union[None, int, float, str] = None

        self.workarounds: Final = workarounds
        self.transformations: Final = transformations
        self.filters: Final = filters

    def set_value(self, sml_value: Optional[SmlListEntry], frame_values: Dict[str, SmlListEntry]):
        self.sml_value = sml_value

        # apply all workarounds
        for workaround in self.workarounds:
            if workaround.enabled:
                sml_value = workaround.fix(sml_value, frame_values)

        # transform the values
        value = None if sml_value is None else sml_value.get_value()
        for f in self.transformations:
            value = f.process(value)
        self.last_value = value

        # check if we want to publish
        do_publish = False
        for refresh in self.filters:
            do_publish = refresh.required(value) or do_publish

        if not do_publish:
            return None

        self.mqtt.publish(value)
        for refresh in self.filters:
            refresh.done(value)

    def describe(self, indent=0, indent_width=2) -> str:

        base = " " * indent
        once = " " * (indent + indent_width)
        twice = " " * (indent + indent_width * 2)

        txt = f'{base}{self.mqtt.topic} ({self.obis}):\n' \
              f'{once}raw value: {self.sml_value.get_value()}\n' \
              f'{once}pub value: {self.last_value}\n'

        if self.workarounds:
            txt += f'{once}workarounds:\n'
            for w in self.workarounds:
                txt += f'{twice}- {w}\n'

        if self.transformations:
            txt += f'{once}transformations:\n'
            for t in self.transformations:
                txt += f'{twice}- {t}\n'

        if self.filters:
            txt += f'{once}filters:\n'
            for f in self.filters:
                txt += f'{twice}- {f}\n'

        return txt



from typing import TypeVar

from sml2mqtt.value.__types__ import VALUE_TYPE


class FilterBase:
    def required(self, value: VALUE_TYPE) -> VALUE_TYPE:
        raise NotImplementedError()

    def done(self, value):
        raise NotImplementedError()


FILTER_OBJ = TypeVar('FILTER_OBJ', bound=FilterBase)



from time import monotonic
from typing import Any, Final, Optional

from .base import FilterBase


class RefreshEvery(FilterBase):
    def __init__(self, refresh_time: float):
        if refresh_time <= 0:
            raise ValueError('Refresh time must be > 0')

        self._refresh_time: Final = refresh_time
        self._last_refresh: Optional[float] = None

        self._last_value: Any = None

    def required(self, value: Any) -> bool:
        if self._last_refresh is None:
            return True

        if monotonic() - self._last_refresh >= self._refresh_time:
            return True

        return False

    def done(self, value: Any):
        self._last_refresh = monotonic()
        self._last_value = value

    def __repr__(self):
        return f'<Every: {self._refresh_time}>'



from sml2mqtt.value.filter.base import FILTER_OBJ

# isort: split

from sml2mqtt.value.filter.change import ChangeFilter
from sml2mqtt.value.filter.diff import DiffAbsFilter, DiffFilterBase, DiffPercFilter
from sml2mqtt.value.filter.time import RefreshEvery



from typing import TypeVar

from sml2mqtt.value.__types__ import VALUE_TYPE


class TransformationBase:
    def process(self, value: VALUE_TYPE) -> VALUE_TYPE:
        raise NotImplementedError()


TRANSFORM_OBJ = TypeVar('TRANSFORM_OBJ', bound=TransformationBase)



from asyncio import sleep
from typing import Optional


class DynDelay:
    def __init__(self, min_delay: float, max_delay: float, start_delay: Optional[float] = None):
        if min_delay < 0:
            raise ValueError(f'min_delay must be >= 0: {min_delay}')
        if max_delay <= min_delay:
            raise ValueError(f'max_delay must be >= min_delay: {max_delay}')

        self.min = min_delay
        self.max = max_delay

        if start_delay is None:
            start_delay = min_delay

        self.curr = max(min(start_delay, max_delay), min_delay)

    async def wait(self):
        await sleep(self.curr)

    def increase(self):
        self.curr = min(self.max, self.curr * 2)
        if not self.curr:
            self.curr = 1

    def reset(self):
        self.curr = self.min

    async def __aenter__(self):
        await self.wait()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.reset()
        else:
            self.increase()
        return False



import time
import traceback
from asyncio import create_task, Future
from typing import Optional, Union

from asyncio_mqtt import Client, MqttError, Will

import sml2mqtt
from sml2mqtt.__log__ import log as _parent_logger

# from sml2mqtt.config import CONFIG
from sml2mqtt.mqtt import DynDelay

log = _parent_logger.getChild('mqtt')

TIME_BEFORE_RECONNECT = 15
DELAY_CONNECT = DynDelay(0, 180)

MQTT: Optional[Client] = None
TASK_CONNECT: Optional[Future] = None


async def disconnect():
    global TASK_CONNECT, MQTT

    if TASK_CONNECT is not None:
        TASK_CONNECT.cancel()
        TASK_CONNECT = None

    if MQTT is not None:
        mqtt = MQTT
        MQTT = None
        if mqtt._client.is_connected():
            await mqtt.disconnect()


async def connect():
    global TASK_CONNECT
    if TASK_CONNECT is None:
        TASK_CONNECT = create_task(_connect_task())


async def _connect_task():
    global TASK_CONNECT
    try:
        await _connect_to_broker()
    finally:
        TASK_CONNECT = None


async def _connect_to_broker():
    global MQTT

    # # We don't publish anything if we just analyze the data from the reader
    # if sml2mqtt._args.ARGS.analyze:
    #     return None

    config = sml2mqtt.config.CONFIG

    while True:
        try:
            async with DELAY_CONNECT:
                # If we are already connected we try to disconnect before we reconnect
                try:
                    if MQTT is not None:
                        if MQTT._client.is_connected():
                            await MQTT.disconnect()
                        MQTT = None
                except Exception as e:
                    log.error(f'Error while disconnecting: {e}')

                # since we just pass this into the mqtt wrapper we do not link it to the base topic
                will_topic = sml2mqtt.mqtt.MqttObj(
                    config.mqtt.topic, config.mqtt.defaults.qos, config.mqtt.defaults.retain
                ).update().create_child(config.mqtt.last_will)
                will_topic.set_config(config.mqtt.last_will)

                MQTT = Client(
                    hostname=config.mqtt.connection.host,
                    port=config.mqtt.connection.port,
                    username=config.mqtt.connection.user if config.mqtt.connection.user else None,
                    password=config.mqtt.connection.password if config.mqtt.connection.password else None,
                    will=Will(will_topic.topic, payload='OFFLINE', qos=will_topic.qos, retain=will_topic.retain)
                )

                log.debug(f'Connecting to {config.mqtt.connection.host}:{config.mqtt.connection.port}')
                await MQTT.connect()
                log.debug('Success!')

                # signal that we are online
                will_topic.publish('ONLINE')
                break

        except MqttError as e:
            log.error(f'{e} ({e.__class__.__name__})')
        except Exception:
            for line in traceback.format_exc().splitlines():
                log.error(line)
            return None


PUBS_FAILED_SINCE: Optional[float] = None


async def publish(topic: str, value: Union[int, float, str], qos: int, retain: bool):
    global PUBS_FAILED_SINCE

    if MQTT is None or not MQTT._client.is_connected():
        await connect()
        return None

    # publish message
    try:
        await MQTT.publish(topic, value, qos=qos, retain=retain)
        PUBS_FAILED_SINCE = None
    except MqttError as e:
        log.error(f'Error while publishing to {topic}: {e} ({e.__class__.__name__})')

        # If we fail too often we try to reconnect
        if PUBS_FAILED_SINCE is None:
            PUBS_FAILED_SINCE = time.time()
        else:
            if time.time() - PUBS_FAILED_SINCE >= TIME_BEFORE_RECONNECT:
                await connect()



from typing import Dict, TypeVar

from smllib.sml import SmlListEntry

from sml2mqtt.value.__types__ import WORKAROUND_TYPE


class WorkaroundBase:
    def __init__(self, arg: WORKAROUND_TYPE):
        self.enabled = True
        if arg is False:
            self.enabled = False

    def fix(self, value: SmlListEntry, frame_values: Dict[str, SmlListEntry]) -> SmlListEntry:
        raise NotImplementedError()

    def __repr__(self):
        return f'<{self.__class__.__name__}>'


WORKAROUND_OBJ = TypeVar('WORKAROUND_OBJ', bound=WorkaroundBase)



from . import __types__, filter

# isort: split

import sml2mqtt.config

# isort: split

from sml2mqtt.value.enum_builder import filter_from_config, transform_from_config, workaround_from_config
from sml2mqtt.value.smlvalue import SmlValue



import asyncio
import platform
import sys
import traceback
import typing

from sml2mqtt.__args__ import get_command_line_args
from sml2mqtt.__log__ import log, setup_log
from sml2mqtt.__shutdown__ import get_return_code, setup_signal_handler, shutdown
from sml2mqtt.config import CONFIG
from sml2mqtt.device import Device
from sml2mqtt.mqtt import BASE_TOPIC, connect


async def a_main():
    await connect()
    await asyncio.sleep(0.1)  # Wait till mqtt is connected

    # Create devices for port
    try:
        for port_cfg in CONFIG.ports:
            dev_mqtt = BASE_TOPIC.create_child(port_cfg.url)
            await Device.create(port_cfg.url, port_cfg.timeout, set(), dev_mqtt)

    except Exception as e:
        shutdown(e)


def main() -> typing.Union[int, str]:
    try:
        CONFIG.load_config_file(get_command_line_args().config)
    except Exception as e:
        print(e)
        return 7

    # This is needed to make async-mqtt work
    if platform.system() == 'Windows':
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    # Add possibility to stop program with Ctrl + c
    setup_signal_handler()
    loop = None

    try:
        setup_log()

        # setup mqtt base topic
        BASE_TOPIC.cfg.topic_fragment = CONFIG.mqtt.topic
        BASE_TOPIC.cfg.qos = CONFIG.mqtt.defaults.qos
        BASE_TOPIC.cfg.retain = CONFIG.mqtt.defaults.retain
        BASE_TOPIC.update()

        # setup loop
        loop = asyncio.get_event_loop_policy().get_event_loop()
        loop.create_task(a_main())
        loop.run_forever()
    except Exception as e:
        for line in traceback.format_exc().splitlines():
            log.error(line)
            print(e)
        return str(e)
    finally:
        if loop is not None:
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    return get_return_code()


if __name__ == "__main__":
    ret = main()
    log.info(f'Closed with return code {ret}')
    sys.exit(ret)



from typing import Any, Dict, List, Type, TypeVar, Union

import sml2mqtt
from sml2mqtt.config.device import FilterOptionEnum, TransformOptionEnum, TYPE_SML_VALUE_FILTER_CFG, \
    TYPE_SML_VALUE_TRANSFORM_CFG, TYPE_SML_VALUE_WORKAROUND_CFG, WorkaroundOptionEnum
from sml2mqtt.value.filter import ChangeFilter, DiffAbsFilter, DiffFilterBase, DiffPercFilter, FILTER_OBJ, RefreshEvery
from sml2mqtt.value.transformations import FactorTransformation, \
    OffsetTransformation, RoundTransformation, TRANSFORM_OBJ
from sml2mqtt.value.workarounds import NegativeOnEnergyMeterStatus, WORKAROUND_OBJ

TYPE_A = TypeVar('TYPE_A')


def _from_config(cfg: Union[TYPE_SML_VALUE_FILTER_CFG, TYPE_SML_VALUE_TRANSFORM_CFG],
                 class_map: Dict[Any, Type[TYPE_A]]) -> List[TYPE_A]:
    if cfg is None:
        return []

    ret = []
    for entry in cfg:
        for key, cls in class_map.items():
            params = entry.get(key)
            if params is not None:
                ret.append(cls(params))
                break
        else:
            raise ValueError(f'Unknown type: {entry}')
    return ret


def filter_from_config(cfg: TYPE_SML_VALUE_FILTER_CFG) -> List[FILTER_OBJ]:
    class_dict = {
        FilterOptionEnum.diff: DiffAbsFilter,
        FilterOptionEnum.perc: DiffPercFilter,
        FilterOptionEnum.every: RefreshEvery,
    }
    filters = _from_config(cfg, class_dict)

    # Default filters
    for f in filters:
        if isinstance(f, RefreshEvery):
            break
    else:
        filters.append(RefreshEvery(sml2mqtt.CONFIG.general.republish_after))

    for f in filters:
        if isinstance(f, DiffFilterBase):
            break
    else:
        filters.append(ChangeFilter())

    return filters


def transform_from_config(cfg: TYPE_SML_VALUE_TRANSFORM_CFG) -> List[TRANSFORM_OBJ]:
    class_dict = {
        TransformOptionEnum.factor: FactorTransformation,
        TransformOptionEnum.round: RoundTransformation,
        TransformOptionEnum.offset: OffsetTransformation,
    }
    return _from_config(cfg, class_dict)


def workaround_from_config(cfg: TYPE_SML_VALUE_WORKAROUND_CFG) -> List[WORKAROUND_OBJ]:
    class_dict = {
        WorkaroundOptionEnum.negative_on_energy_meter_status: NegativeOnEnergyMeterStatus,
    }
    return _from_config(cfg, class_dict)



from asyncio import create_task, Event, Task, TimeoutError, wait_for
from typing import Any, Callable, Final, Optional


class Watchdog:
    def __init__(self, timeout: float, callback: Callable[[], Any]):
        if timeout <= 0:
            raise ValueError()
        self.timeout: Final = timeout
        self.callback: Final = callback
        self.event: Final = Event()
        self.task: Optional[Task] = None

    def start(self):
        if self.task is None:
            self.task = create_task(self.wd_task())

    def cancel(self):
        if self.task is not None:
            self.task.cancel()
            self.task = None

    def feed(self):
        self.event.set()

    async def wd_task(self):
        try:
            make_call = True
            while True:
                self.event.clear()

                try:
                    await wait_for(self.event.wait(), self.timeout)
                    make_call = True
                    continue
                except TimeoutError:
                    pass

                # callback only once!
                if make_call:
                    make_call = False
                    self.callback()
        finally:
            self.task = None



import dataclasses
from asyncio import create_task
from typing import List, Optional, Union

from sml2mqtt import CMD_ARGS
from sml2mqtt.__log__ import get_logger
from sml2mqtt.config import OptionalMqttPublishConfig
from sml2mqtt.mqtt import publish


class TopicFragmentExpected(Exception):
    pass


class MqttConfigValuesMissing(Exception):
    pass


@dataclasses.dataclass
class MqttCfg:
    topic_full: Optional[str] = None
    topic_fragment: Optional[str] = None
    qos: Optional[int] = None
    retain: Optional[bool] = None

    def is_full_config(self) -> bool:
        if self.topic_fragment is None and self.topic_full is None:
            return False
        if self.qos is None:
            return False
        if self.retain is None:
            return False
        return True


class MqttObj:
    def __init__(self, topic_fragment: Optional[str] = None, qos: Optional[int] = None, retain: Optional[bool] = None):

        # Configured parts
        self.cfg = MqttCfg(topic_fragment=topic_fragment, qos=qos, retain=retain)

        # Effective config
        self.qos: int = 0
        self.retain: bool = False
        self.topic: str = ''

        self.parent: Optional[MqttObj] = None
        self.children: List[MqttObj] = []

    def publish(self, value: Union[str, int, float, bytes]):
        # do not publish when the analyze flag is set
        if CMD_ARGS.analyze:
            get_logger('mqtt.pub').info(f'{self.topic}: {value} (QOS: {self.qos}, retain: {self.retain})')
        else:
            create_task(publish(self.topic, value, self.qos, self.retain))

    def update(self) -> 'MqttObj':
        self._merge_values()
        for c in self.children:
            c.update()
        return self

    def _merge_values(self) -> 'MqttObj':

        if self.parent is None:
            if not self.cfg.is_full_config():
                raise MqttConfigValuesMissing()

        if self.cfg.topic_full:
            self.topic = self.cfg.topic_full
        else:
            if not self.cfg.topic_fragment:
                raise TopicFragmentExpected()
            if self.parent is None:
                self.topic = self.cfg.topic_fragment
            else:
                self.topic = f'{self.parent.topic}/{self.cfg.topic_fragment}'

        self.qos = self.cfg.qos
        if self.qos is None:
            self.qos = self.parent.qos

        self.retain = self.cfg.retain
        if self.retain is None:
            self.retain = self.parent.retain
        return self

    def set_topic(self, topic: str) -> 'MqttObj':
        self.cfg.topic_fragment = topic
        self.update()
        return self

    def set_config(self, cfg: Optional[OptionalMqttPublishConfig]) -> 'MqttObj':
        if cfg is None:
            return self

        local = self.cfg
        local.topic_full = cfg.full_topic
        local.topic_fragment = cfg.topic
        local.qos = cfg.qos
        local.retain = cfg.retain
        self.update()
        return self

    def create_child(self, topic_fragment: Optional[str] = None, qos: Optional[int] = None,
                     retain: Optional[bool] = None) -> 'MqttObj':
        child = self.__class__(topic_fragment=topic_fragment, qos=qos, retain=retain)
        child.parent = self
        self.children.append(child)
        child.update()
        return child


BASE_TOPIC = MqttObj()



from typing import Dict, List, Union

from easyconfig import AppBaseModel, BaseModel, create_app_config
from pydantic import constr, Field

from .device import REPUBLISH_ALIAS, SmlDeviceConfig, SmlValueConfig
from .logging import LoggingSettings
from .mqtt import MqttConfig, OptionalMqttPublishConfig


class PortSettings(BaseModel):
    url: constr(strip_whitespace=True, min_length=1, strict=True) = Field(..., description='Device path')
    timeout: Union[int, float] = Field(
        default=3, description='Seconds after which a timeout will be detected (default=3)')


class GeneralSettings(BaseModel):
    wh_in_kwh: bool = Field(True, description='Automatically convert Wh to kWh', alias='Wh in kWh')
    republish_after: int = Field(
        120, description='Republish automatically after this time (if no other filter configured)',
        alias=REPUBLISH_ALIAS,
    )
    report_blank_energy_meters: bool = Field(
        False, description='Report blank energy meters (where the value is 0kwh)',
        alias='report blank energy meters', in_file=False
    )
    report_device_id: bool = Field(
        False, description='Report the device id even though it does never change',
        alias='report device id', in_file=False
    )
    device_id_obis: str = Field(
        '', description='Additional OBIS field for the serial number, default is 0100000009ff',
        alias='device id obis', in_file=False
    )


class Settings(AppBaseModel):
    logging: LoggingSettings = LoggingSettings()
    mqtt: MqttConfig = MqttConfig()
    general: GeneralSettings = GeneralSettings()
    ports: List[PortSettings] = []
    devices: Dict[str, SmlDeviceConfig] = Field({}, description='Device configuration by ID or url',)


def default_config() -> Settings:
    # File defaults
    s = Settings(
        ports=[PortSettings(url='COM1', timeout=3), PortSettings(url='/dev/ttyS0', timeout=3), ],
        devices={
            'DEVICE_ID_HEX': SmlDeviceConfig(
                mqtt=OptionalMqttPublishConfig(topic='DEVICE_BASE_TOPIC'),
                status=OptionalMqttPublishConfig(topic='status'),
                skip=['OBIS'],
                values={
                    'OBIS': SmlValueConfig(
                        mqtt=OptionalMqttPublishConfig(topic='OBIS'),
                        workarounds=[{'negative on energy meter status': True}],
                        transformations=[{'factor': 3}, {'offset': 100}, {'round': 2}],
                        filters=[{'diff': 10}, {'perc': 10}, {'every': 120}],
                    )
                }
            )
        }
    )
    return s


CONFIG = create_app_config(Settings(), default_config)



from typing import Final, Union

from .base import FilterBase, VALUE_TYPE


class DiffFilterBase(FilterBase):
    def __init__(self, min_diff: Union[int, float]):
        if min_diff < 0:
            raise ValueError('min diff must be >= 0')
        self.min_diff: Final = min_diff
        self.last_value: VALUE_TYPE = None

    def required(self, value: VALUE_TYPE) -> VALUE_TYPE:
        if value is None:
            return False
        if self.last_value is None:
            return True
        return self.diff(value)

    def done(self, value):
        self.last_value = value

    def diff(self, value: Union[int, float]) -> VALUE_TYPE:
        raise NotImplementedError()


class DiffAbsFilter(DiffFilterBase):
    def diff(self, value: Union[int, float]) -> VALUE_TYPE:
        if abs(value - self.last_value) < self.min_diff:
            return False
        return True

    def __repr__(self):
        return f'<AbsDiff: {self.min_diff}>'


class DiffPercFilter(DiffFilterBase):
    def diff(self, value: Union[int, float]) -> VALUE_TYPE:
        if not self.last_value:
            return False

        perc = abs(1 - value / self.last_value) * 100
        if perc < self.min_diff:
            return False
        return True

    def __repr__(self):
        return f'<PercDiff: {self.min_diff}>'



import argparse
import os
import sys
from pathlib import Path
from typing import Final, Optional, Type


class CommandArgs:
    config: Optional[Path] = None
    analyze: bool = False


CMD_ARGS: Final = CommandArgs


def get_command_line_args(args=None) -> Type[CommandArgs]:

    parser = argparse.ArgumentParser(description='SML to MQTT bridge')
    parser.add_argument(
        '-c',
        '--config',
        help='Path to configuration file',
        default=None
    )
    parser.add_argument(
        '-a',
        '--analyze',
        help='Processes exactly one sml message, shows the values of the message and what will be reported',
        action='store_true',
        default=False
    )
    args = parser.parse_args(args)
    CMD_ARGS.config = find_config_folder(args.config)
    CMD_ARGS.analyze = args.analyze
    return CMD_ARGS


def find_config_folder(config_file_str: Optional[str]) -> Path:
    check_path = []

    if config_file_str is None:
        # Nothing is specified, we try to find the config automatically
        try:
            working_dir = Path(os.getcwd())
            check_path.append(working_dir / 'sml2mqtt')
            check_path.append(working_dir.with_name('sml2mqtt'))
            check_path.append(working_dir.parent.with_name('sml2mqtt'))
        except ValueError:
            # the ValueError gets raised if the working_dir or its parent is empty (e.g. c:\ or /)
            pass

        check_path.append(Path.home() / 'sml2mqtt')   # User home

        # if we run in a venv check the venv, too
        if v_env := os.environ.get('VIRTUAL_ENV', ''):
            check_path.append(Path(v_env) / 'sml2mqtt')  # Virtual env dir

        for i, config_file in enumerate(check_path):
            check_path[i] = config_file = config_file.resolve() / 'config.yml'
            if config_file.is_file():
                return config_file
    else:
        config_file = Path(config_file_str).resolve()
        # Add to check_path so we have a nice error message
        check_path.append(config_file)

        if config_file.is_file():
            return config_file

        # folder exists but not the file -> create
        if config_file.parent.is_dir():
            return config_file

    # we have nothing found and nothing specified -> exit
    print('Config file not found!')
    print('Checked paths:\n - ' + '\n - '.join(str(k) for k in check_path))
    print('Please create file or specify a path with the "-c" arg switch.')
    sys.exit(1)



from easyconfig import BaseModel
from pydantic import conint, constr, Field, StrictBool, validator

QOS = conint(ge=0, le=2)
TOPIC_STR = constr(strip_whitespace=True, min_length=1)
STRIPPED_STR = constr(strip_whitespace=True)


class MqttDefaultPublishConfig(BaseModel):
    qos: QOS = Field(
        0, description='Default value for QOS if no other QOS value in the config entry is set')
    retain: StrictBool = Field(
        False, description='Default value for retain if no other retain value in the config entry is set')


class OptionalMqttPublishConfig(BaseModel):
    topic: TOPIC_STR = None
    full_topic: TOPIC_STR = Field(None, alias='full topic')
    qos: QOS = None
    retain: StrictBool = None

    @validator('topic', 'full_topic')
    def validate_topic(cls, value):
        if value is None:
            return None

        if value.endswith('/'):
            raise ValueError('Topic must not end with "/"')
        if value.startswith('/'):
            raise ValueError('Topic must not start with "/"')
        return value

    @validator('full_topic')
    def check_full_or_partial(cls, v, values):
        if v is None:
            return None

        if values.get('topic') is not None:
            raise ValueError('Topic and full_topic can not be used at the same time!')
        return v


class MqttConnection(BaseModel):
    client_id: STRIPPED_STR = Field('sml2mqtt', alias='client id')
    host: STRIPPED_STR = 'localhost'
    port: conint(gt=0) = 1883
    user: STRIPPED_STR = ''
    password: STRIPPED_STR = ''
    tls: StrictBool = False
    tls_insecure: StrictBool = Field(False, alias='tls insecure')


class MqttConfig(BaseModel):
    connection: MqttConnection = MqttConnection()
    topic: TOPIC_STR = Field('sml2mqtt', alias='topic prefix')
    defaults: MqttDefaultPublishConfig = MqttDefaultPublishConfig()
    last_will: OptionalMqttPublishConfig = Field(
        default_factory=lambda: OptionalMqttPublishConfig(topic='status'), alias='last will')



import signal
import traceback
from asyncio import create_task, get_event_loop, sleep, Task
from typing import Dict, Optional, Type, Union

import sml2mqtt.mqtt
from sml2mqtt.__log__ import log
from sml2mqtt.errors import AllDevicesFailed, DeviceSetupFailed

# ----------------------------------------------------------------------------------------------------------------------
# Return code logic
# ----------------------------------------------------------------------------------------------------------------------
_RETURN_CODE: Optional[int] = None


def set_return_code(code: int):
    global _RETURN_CODE
    if _RETURN_CODE is None:
        _RETURN_CODE = code
    else:
        if _RETURN_CODE != code:
            log.debug(f'Return code is already set to {_RETURN_CODE}, skip setting {code}!')


def get_return_code() -> int:
    if _RETURN_CODE is None:
        log.warning('No return code set!')
        return 2
    return _RETURN_CODE


# ----------------------------------------------------------------------------------------------------------------------
# Signal handlers so we can shutdown gracefully
# ----------------------------------------------------------------------------------------------------------------------
def shutdown_handler(sig, frame):
    set_return_code(0)
    create_task(do_shutdown())


def setup_signal_handler():
    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)


# ----------------------------------------------------------------------------------------------------------------------
# Actual shutdown logic
# ----------------------------------------------------------------------------------------------------------------------
SHUTDOWN_TASK: Optional[Task] = None


async def do_shutdown():
    global SHUTDOWN_TASK

    try:
        print('Shutting down ...')
        log.info('Shutting down ...')

        for device in sml2mqtt.device.sml_device.ALL_DEVICES.values():
            device.shutdown()

        await sml2mqtt.mqtt.disconnect()
        await sleep(0.1)

        get_event_loop().stop()
    finally:
        SHUTDOWN_TASK = None


# ----------------------------------------------------------------------------------------------------------------------
# shutdown helper
# ----------------------------------------------------------------------------------------------------------------------
def shutdown(e: Union[Exception, Type[Exception]]):
    global SHUTDOWN_TASK

    ret_map: Dict[int, Type[Exception]] = {10: DeviceSetupFailed, 20: AllDevicesFailed}

    log_traceback = True

    # get return code based on the error
    for ret_code, cls in ret_map.items():
        if isinstance(e, cls):
            break

        if e is cls:
            log_traceback = False
            break
    else:
        ret_code = 1

    if log_traceback:
        for line in traceback.format_exc().splitlines():
            log.error(line)

    set_return_code(ret_code)

    if SHUTDOWN_TASK is None:
        SHUTDOWN_TASK = create_task(do_shutdown())



from sml2mqtt import __log__
from sml2mqtt.__args__ import CMD_ARGS
from sml2mqtt.__version__ import __version__

# isort: split

from sml2mqtt.config import CONFIG

# isort: split

from sml2mqtt import value
from sml2mqtt.process_value import process_value

# isort: split

import sml2mqtt.device



import asyncio
from asyncio import create_task, Task
from typing import Awaitable, Callable, Final, Optional, TYPE_CHECKING

from serial_asyncio import create_serial_connection, SerialTransport

from sml2mqtt.__log__ import get_logger
from sml2mqtt.device import DeviceStatus

if TYPE_CHECKING:
    import sml2mqtt


log = get_logger('serial')


class SmlSerial(asyncio.Protocol):
    @classmethod
    async def create(cls, url: str, device: 'sml2mqtt.device_id.Device') -> 'SmlSerial':
        transport, protocol = await create_serial_connection(
            asyncio.get_event_loop(), cls, url, baudrate=9600)  # type: SerialTransport, SmlSerial

        protocol.url = url
        protocol.device = device
        return protocol

    def __init__(self) -> None:
        super().__init__()

        self.url: Optional[str] = None
        self.device: Optional['sml2mqtt.device_id.Device'] = None

        self.transport: Optional[SerialTransport] = None

        self.pause_task: Optional[Task] = None

        self.on_data_cb: Final = Callable[[bytes], Awaitable]

    def connection_made(self, transport):
        self.transport = transport
        log.debug(f'Port {self.url} successfully opened')

        self.device.set_status(DeviceStatus.PORT_OPENED)

    def data_received(self, data: bytes):
        self.pause_serial()
        create_task(self.device.serial_data_read(data))

    async def resume_serial(self):
        try:
            await asyncio.sleep(0.4)
            self.transport.resume_reading()
        finally:
            self.pause_task = None

    def pause_serial(self):
        self.transport.pause_reading()
        self.pause_task = create_task(self.resume_serial())

    def connection_lost(self, exc):
        self.close()

        log.info(f'Port {self.url} was closed')
        self.device.set_status(DeviceStatus.PORT_CLOSED)

    def close(self):
        if self.pause_task is not None:
            self.pause_task.cancel()
            self.pause_task = None

        if not self.transport.is_closing():
            self.transport.close()



from enum import Enum
from typing import Dict, List, Optional, Set, Union

from easyconfig import BaseModel
from pydantic import Field, StrictBool, StrictFloat, StrictInt, StrictStr, validator

from .mqtt import OptionalMqttPublishConfig

REPUBLISH_ALIAS = 'republish after'


class WorkaroundOptionEnum(str, Enum):
    negative_on_energy_meter_status = 'negative on energy meter status'


class TransformOptionEnum(str, Enum):
    factor = 'factor'
    offset = 'offset'
    round = 'round'


class FilterOptionEnum(str, Enum):
    diff = 'diff'
    perc = 'perc'
    every = 'every'


TYPE_SML_VALUE_WORKAROUND_CFG = \
    Optional[List[Dict[WorkaroundOptionEnum, Union[StrictBool, StrictInt, StrictFloat, StrictStr]]]]
TYPE_SML_VALUE_TRANSFORM_CFG = Optional[List[Dict[TransformOptionEnum, Union[StrictInt, StrictFloat]]]]
TYPE_SML_VALUE_FILTER_CFG = Optional[List[Dict[FilterOptionEnum, Union[StrictInt, StrictFloat]]]]


class SmlValueConfig(BaseModel):
    mqtt: OptionalMqttPublishConfig = Field(None, description='Mqtt config for this entry (optional)')

    workarounds: TYPE_SML_VALUE_WORKAROUND_CFG = Field(
        None, description='Workarounds for the value (optional)')
    transformations: TYPE_SML_VALUE_TRANSFORM_CFG = Field(
        None, description='Mathematical transformations for the value (optional)')
    filters: TYPE_SML_VALUE_FILTER_CFG = Field(
        None, description='Refresh options for the value (optional)')

    @validator('workarounds', 'transformations', 'filters')
    def len_1(cls, v):
        if v is None:
            return None

        for entry in v:
            if len(entry) != 1:
                raise ValueError(f'Only one entry allowed! Got {len(entry)}: {", ".join(entry.keys())}')
        return v


class SmlDeviceConfig(BaseModel):
    mqtt: Optional[OptionalMqttPublishConfig] = OptionalMqttPublishConfig()
    status: Optional[OptionalMqttPublishConfig] = OptionalMqttPublishConfig(topic='status')

    skip: Optional[Set[StrictStr]] = Field(
        default=None, description='OBIS codes (HEX) of values that will not be published (optional)')

    values: Optional[Dict[str, SmlValueConfig]] = Field(
        default=None, description='Special configurations for each of the values (optional)')



from typing import Final, Union

from .base import TransformationBase, VALUE_TYPE


class FactorTransformation(TransformationBase):
    def __init__(self, factor: Union[int, float]):
        self.factor: Final = factor

    def process(self, value: VALUE_TYPE) -> VALUE_TYPE:
        if value is None:
            return None
        return value * self.factor

    def __repr__(self):
        return f'<Factor: {self.factor}>'


class OffsetTransformation(TransformationBase):
    def __init__(self, offset: Union[int, float]):
        self.offset: Final = offset

    def process(self, value: VALUE_TYPE) -> VALUE_TYPE:
        if value is None:
            return None
        return value + self.offset

    def __repr__(self):
        return f'<Offset: {self.offset}>'


class RoundTransformation(TransformationBase):
    def __init__(self, digits: int):
        self.digits: Final = digits if digits else None

    def process(self, value: VALUE_TYPE) -> VALUE_TYPE:
        if value is None:
            return None
        if isinstance(value, int):
            return value
        return round(value, self.digits)

    def __repr__(self):
        return f'<Round: {self.digits if self.digits is not None else 0}>'



from typing import Dict

from smllib.sml import SmlListEntry

from .base import WORKAROUND_TYPE, WorkaroundBase


class NegativeOnEnergyMeterStatus(WorkaroundBase):
    def __init__(self, arg: WORKAROUND_TYPE):
        super().__init__(arg)

        self.meter_obis = '0100010800ff'
        if isinstance(arg, str):
            self.meter_obis = arg

    def fix(self, value: SmlListEntry, frame_values: Dict[str, SmlListEntry]) -> SmlListEntry:
        meter = frame_values.get(self.meter_obis)
        if meter is None:
            raise ValueError(f'Configured meter obis "{self.meter_obis}" not found in current frame')

        status = meter.status
        if not isinstance(status, int):
            raise ValueError(f'Energy Meter status is not a valid int: {status} ({type(status)})')

        negative = status & 0x20
        if negative:
            value.value *= -1



from enum import Enum


class DeviceStatus(str, Enum):
    STARTUP = 'STARTUP'
    SHUTDOWN = 'SHUTDOWN'
    PORT_OPENED = 'PORT_OPENED'
    PORT_CLOSED = 'PORT_CLOSED'
    MSG_TIMEOUT = 'MSG_TIMEOUT'
    CRC_ERROR = 'CRC_ERROR'
    ERROR = 'ERROR'
    OK = 'OK'

    def is_shutdown_status(self) -> bool:
        return self.value in (DeviceStatus.PORT_CLOSED, DeviceStatus.ERROR, DeviceStatus.SHUTDOWN)



import logging
from typing import Dict

from smllib.sml import SmlListEntry

import sml2mqtt
from sml2mqtt.__args__ import CMD_ARGS
from sml2mqtt.__log__ import get_logger
from sml2mqtt.value import SmlValue

VALUES: Dict[str, Dict[str, SmlValue]] = {}


values_log = get_logger('values')


def process_value(device: 'sml2mqtt.device.Device', obj: SmlListEntry, frame_values: Dict[str, SmlListEntry]):
    try:
        device_values = VALUES[device.device_id]
    except KeyError:
        VALUES[device.device_id] = device_values = {}

    try:
        value = device_values[str(obj.obis)]
    except KeyError:
        device_values[str(obj.obis)] = value = create_sml_value(device, obj)

    value.set_value(obj, frame_values)


def create_sml_value(device: 'sml2mqtt.device.Device', obj: SmlListEntry) -> SmlValue:
    obis = str(obj.obis)
    device_id = device.device_id
    log_level = logging.DEBUG if not CMD_ARGS.analyze else logging.INFO

    cfg = None
    if sml2mqtt.CONFIG.devices is not None:
        cfg = sml2mqtt.CONFIG.devices.get(device_id)

    def create_default_value():
        values_log.log(log_level, f'Creating default value handler for {device_id}/{obis}')
        return SmlValue(
            device_id, obis, device.mqtt_device.create_child(obis),
            workarounds=[],
            transformations=[],
            filters=sml2mqtt.value.filter_from_config(None)
        )

    if cfg is None or cfg.values is None:
        return create_default_value()

    value_cfg = cfg.values.get(obis)
    if value_cfg is None:
        return create_default_value()

    values_log.log(log_level, f'Creating value handler from config for {device_id}/{obis}')

    return SmlValue(
        device_id, obis, device.mqtt_device.create_child(obis).set_config(value_cfg.mqtt),
        workarounds=sml2mqtt.value.workaround_from_config(value_cfg.workarounds),
        transformations=sml2mqtt.value.transform_from_config(value_cfg.transformations),
        filters=sml2mqtt.value.filter_from_config(value_cfg.filters),
    )



import traceback
from binascii import b2a_hex
from typing import Dict, Final, List, Optional, Set

from smllib import SmlFrame, SmlStreamReader
from smllib.errors import CrcError
from smllib.sml import SmlListEntry

import sml2mqtt
from sml2mqtt import process_value
from sml2mqtt.__log__ import get_logger
from sml2mqtt.__shutdown__ import shutdown
from sml2mqtt.config import CONFIG
from sml2mqtt.config.device import SmlDeviceConfig
from sml2mqtt.device import DeviceStatus
from sml2mqtt.device.watchdog import Watchdog
from sml2mqtt.errors import AllDevicesFailed, DeviceSetupFailed
from sml2mqtt.mqtt import MqttObj
from sml2mqtt.process_value import VALUES as ALL_VALUES

ALL_DEVICES: Dict[str, 'Device'] = {}


class Device:
    @classmethod
    async def create(cls, url: str, timeout: float, skip_values: Set[str], mqtt_device: MqttObj):
        try:
            device = cls(url, timeout, set(skip_values), mqtt_device)
            device.serial = await sml2mqtt.device.SmlSerial.create(url, device)
            ALL_DEVICES[url] = device
            return device
        except Exception as e:
            raise DeviceSetupFailed(e) from None

    def __init__(self, url: str, timeout: float, skip_values: Set[str], mqtt_device: MqttObj):
        self.stream = SmlStreamReader()
        self.serial: 'sml2mqtt.device.SmlSerial' = None
        self.watchdog: Final = Watchdog(timeout, self.serial_data_timeout)

        self.status: DeviceStatus = DeviceStatus.STARTUP
        self.mqtt_device: MqttObj = mqtt_device
        self.mqtt_status: MqttObj = mqtt_device.create_child('status')

        self.log = get_logger(url.split("/")[-1])
        self.log_status = get_logger(url.split("/")[-1]).getChild('status')

        self.device_url = url
        self.device_id: str = url.split("/")[-1]
        self.device_id_set = False

        self.skip_values = skip_values

    def shutdown(self):
        if not self.status.is_shutdown_status():
            self.set_status(DeviceStatus.SHUTDOWN)

    def set_status(self, new_status: DeviceStatus) -> bool:
        if self.status == new_status:
            return False

        self.status = new_status
        self.log_status.info(f'{new_status:s}')

        # Don't publish the port open because we don't have the correct name yet
        if new_status != DeviceStatus.PORT_OPENED:
            self.mqtt_status.publish(new_status.value)

        # If all ports are closed, or we have errors we shut down
        if all(map(lambda x: x.status.is_shutdown_status(), ALL_DEVICES.values())):
            # Stop reading from the serial port because we are shutting down
            self.serial.close()
            shutdown(AllDevicesFailed)
        return True

    def set_device_id(self, serial: str):
        assert isinstance(serial, str)
        assert not self.device_id_set
        self.device_id = serial
        self.device_id_set = True

        # config is optional
        cfg: Optional[SmlDeviceConfig] = None
        if CONFIG.devices is not None:
            cfg = CONFIG.devices.get(serial)

        # Build the defaults but with the serial number instead of the url
        if cfg is None:
            self.mqtt_device.set_topic(serial)
            return None

        # config found -> load all configured values
        self.mqtt_device.set_config(cfg.mqtt)
        self.mqtt_status.set_config(cfg.status)

        if cfg.skip is not None:
            self.skip_values.update(cfg.skip)

    def select_device_id(self, frame_values: Dict[str, SmlListEntry]):
        obis_ids = ['0100000009ff']
        if CONFIG.general.device_id_obis:
            obis_ids.append(CONFIG.general.device_id_obis)

        for obis_id in obis_ids:
            entry = frame_values.pop(obis_id, None)
            if entry is not None:
                if not CONFIG.general.report_device_id:
                    self.skip_values.add(obis_id)
                self.set_device_id(entry.get_value())
                break
        else:
            self.set_device_id(self.device_url)

        # remove additionally skipped frames
        # this gets filled in set_device_id, so we have to do it afterwards!
        for name in self.skip_values:
            frame_values.pop(name, None)

    def serial_data_timeout(self):
        if self.set_status(DeviceStatus.MSG_TIMEOUT):
            self.stream.clear()
            self.log.warning('Timeout')

    async def serial_data_read(self, data: bytes):
        frame = None

        try:
            self.watchdog.feed()
            self.stream.add(data)

            try:
                frame = self.stream.get_frame()
                if frame is None:
                    return None
            except CrcError as e:
                self.log.error(f'Crc error: {e.crc_calc} != {e.crc_msg}')
                self.set_status(DeviceStatus.CRC_ERROR)
                return None

            # Process Frame
            await self.process_frame(frame)
        except Exception:
            # dump frame if possible
            if frame is not None:
                self.log.error('Received Frame')
                self.log.error(f' -> {b2a_hex(frame.buffer)}')

            # Log exception
            for line in traceback.format_exc().splitlines():
                self.log.error(line)

            # Signal that an error occurred
            self.set_status(DeviceStatus.ERROR)
            return None

    async def process_frame(self, frame: SmlFrame):

        do_analyze = sml2mqtt.CMD_ARGS.analyze
        do_wh_in_kwh = CONFIG.general.wh_in_kwh
        report_blank_meters = CONFIG.general.report_blank_energy_meters

        if do_analyze:
            self.log.info('')
            self.log.info('Received Frame')
            self.log.info(f' -> {b2a_hex(frame.buffer)}')
            self.log.info('')
            for obj in frame.parse_frame():
                for line in obj.format_msg().splitlines():
                    self.log.info(line)
            self.log.info('')

        # try shortcut, if that fails try parsing the whole frame
        try:
            sml_objs: List[SmlListEntry] = frame.get_obis()
        except Exception:
            self.log.info('get_obis failed - try parsing frame')
            for line in traceback.format_exc().splitlines():
                self.log.debug(line)

            sml_objs: List[SmlListEntry] = []
            for msg in frame.parse_frame():
                for val in getattr(msg.message_body, 'val_list', []):
                    sml_objs.append(val)

        frame_values: Dict[str, SmlListEntry] = {}
        for sml_obj in sml_objs:
            name = sml_obj.obis
            if name in self.skip_values:
                continue

            # Unit is Wh -> Value is from an Energy Meter
            if sml_obj.unit == 30:
                # We Don't publish disabled energy meters (always 0)
                if sml_obj.get_value() < 0.1 and not report_blank_meters:
                    continue
                # Global option to automatically convert from Wh to kWh
                if do_wh_in_kwh:
                    sml_obj.value /= 1000

            # Mark for publishing
            frame_values[name] = sml_obj

        # We overwrite the device_id url (default) with the serial number if the device reports it
        # Otherwise we still do the config lookup so the user can configure the mqtt topics
        if not self.device_id_set:
            self.select_device_id(frame_values)

        # Process all values
        for obis_value in frame_values.values():
            process_value(self, obis_value, frame_values)

        # There was no Error -> OK
        self.set_status(DeviceStatus.OK)

        if do_analyze:
            for obis, value in ALL_VALUES.get(self.device_id, {}).items():
                self.log.info('')
                for line in value.describe().splitlines():
                    self.log.info(line)

            self.log.info('')
            self.set_status(DeviceStatus.SHUTDOWN)
            return None



from .base import FilterBase, VALUE_TYPE


class ChangeFilter(FilterBase):
    def __init__(self):
        self.last_value: VALUE_TYPE = None

    def required(self, value: VALUE_TYPE) -> VALUE_TYPE:
        if self.last_value is None:
            return True

        if value != self.last_value:
            return True

        return False

    def done(self, value):
        self.last_value = value

    def __repr__(self):
        return '<OnChange>'



import logging
from pathlib import Path

from easyconfig import BaseModel
from pydantic import Extra, Field, validator


class LoggingSettings(BaseModel):
    level: str = Field('INFO', description='Log level')
    file: Path = Field('sml2mqtt.log', description='Log file path (absolute or relative to config file)')

    class Config:
        extra = Extra.forbid

    @validator('level')
    def validate_logging(cls, value):
        if value not in logging._nameToLevel:
            raise ValueError(f'Level must be one of {", ".join(logging._nameToLevel)}')
        return value

    def set_log_level(self) -> int:
        level = logging._nameToLevel[self.level]
        logging.getLogger().setLevel(level)
        return level
